/* @source sigplot application
**
** Generates gnuplot data files of signature performance
** @author: Copyright (C) Matt Blades (mblades@hgmp.mrc.ac.uk)
** @author: Copyright (C) Jon Ison (jison@hgmp.mrc.ac.uk)
** @@
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************
** 
** 
** 
** 
** 
** 
** Operation
** 
** Sigplot parses a signature hits output file generated by the EMBOSS 
** application sigscan, and produces data files suitable for use in the 
** graph plotting application GNUPLOT.  Sigplot parses the secondary 
** classification column of the signature hits output file (Figure 1).
** The data files produced by sigplot when plotted using GNUPLOT, display 
** a single graph indicating the performance of the signature.  The x axis 
** of the graph shows the number of hits and the y axis shows the proportion 
** of the hits which are 'TRUE', 'CROSS', 'FALSE' and 'UNKNOWN', see 
** classification of hits (Figure 1).
**
**
** Excerpts from a signature hits (Figure 1) are shown. The records used 
** are are as follows
** (1) DE - bibliographic information. The text 'Results of signature search'
** is always given.
** Four SCOP classification records are given:
** (2)  CL - Domain class.  It is identical to the text given after 'Class' in 
** the scop classification file (see documentation for the EMBOSS application 
** scope).
** (3)  FO - Domain fold.  It is identical to the text given after 'Fold' in 
** the scop classification file (see scope documentation).
** (4)  SF - Domain superfamily.  It is identical to the text given after 
** 'Superfamily' in the scop classification file (see scope documentation).
** (5)  FA - Domain family. It is identical to the text given after 'Family' in 
** the scop classification file (see scope documentation).
** (6) HI - hit data.  The data are as follows (column numbers are given in 
** parentheses). (i) HI is always given. (ii) Rank-order of the hit. (iii) 
** Database identifier code. (iv) The group number of the protein if a grouped 

** scop families file was provided or '.' otherwise. (v) The primary
** classification of the hit. For true hits (genuine relatives to the signature)
** one of 'TRAIN', 'PSIBLAST' or 'OTHER' is given. Otherwise, 'CROSS', 'FALSE'
** or 'UNKNOWN' is given ('.' is given if a scop families file was not 
** provided). (vi) The secondary classification of the hit, either 'FALSE', 
** 'TRUE' or 'UNKNOWN' ('.' is given if a scop families file was not 
** provided). The classes of hits are defined below.  
** (vii) Score of sequence-signature match.  (viii) E-value (see below).
** (7) XX - used for spacing.
** (8) // - The file ends with a line containing '//' only.
** 
** 
** Figure 1   Excerpt from a signature hits file
** 
** DE   Results of signature search
** XX
** CL   All alpha proteins
** XX
** FO   Globin-like
** XX
** SF   Globin-like
** XX
** FA   Globins
** XX

** XX
** HI   1    1RBPDFG   1    TRUE     TRUE    234  0.0001 
** HI   2    1GFT35J   3    TRUE     TRUE    234  0.0008 
** HI   3    1KJUFGH   1    TRUE     TRUE    224  0.0108 
** HI   4    1GYU15R   2    CLOSE    TRUE    220  0.1876 
** HI   5    1LKI89O   2    CLOSE    TRUE    203  0.6787 
** HI   6    1QRTY58   1    TRUE     TRUE    199  0.9978 
** HI   7    2IOM78G   1    FALSE    FALSE   198  1.0844
** HI   8    1SZR234   1    CLOSE    TRUE    198  1.4343 
** HI   9    3PONI57   1    FALSE    FALSE   197  2.8849 
** HI  10    1PHDJBS   3    CLOSE    TRUE    190  2.9872
** HI  11    1HIOHDW   1    UNKNOWN  UNKNOWN 160  5,8676 
** HI  12    199976T   1    CLOSE    TRUE    140  8.8346 
** XX

** //
** 
** 
**  Notes
**  MJB - ADD IN FORMAT OF GNUPLOT DATA FILES HERE!!!!
******************************************************************************/


#include "emboss.h"
#include <math.h>




typedef struct AjSScopdata
{
    AjPStr   Class;
    AjPStr   Fold;
    AjPStr   Superfamily;
    AjPStr   Family;
    AjPStr   file_name;
    ajint    num_hits;            /* No. of hits */
} AjOScopdata, *AjPScopdata;







static AjBool   sigplot_CountHits(AjPFile hitsin, AjPScopdata *data);
static AjBool  sigplot_HitProportion(AjPFile hitsin, AjPScopdata *data,
				     AjPFloat2d *prob_array);
static AjBool  sigplot_DataWrite(AjPFile datafile, AjPScopdata data,
				 AjPFloat2d prob_array, 
				 AjPStr true, AjPStr false, AjPStr cross,
				 AjPStr unknown);
static void sigplot_ScopdataDel(AjPScopdata *pthis);
static AjPScopdata  sigplot_ScopdataNew();









/* @prog sigplot *************************************************************
**
** Signature performance plotting program
**
******************************************************************************/


int main(int argc, char **argv)
{

    AjPFile     hitsin      = NULL;      /* Signature hits input file */
    AjPFile     datafile    = NULL;      /* GNUPLOT output data file */
    AjPFloat2d  prob_array  = NULL;      /* Array for probabilities of each classification */    
    AjPScopdata data        = NULL;      /* Pointer to Scopdata structure */
    AjPStr      true   = NULL;      /* Name of sigplot data file of true hits */
    AjPStr      false  = NULL;      /* Name of sigplot data file of false hits */
    AjPStr      cross  = NULL;      /* Name of sigplot data file of cross hits */
    AjPStr      unknown= NULL;      /* Name of sigplot data file of unknown hits */
    
    
    embInit("sigplot", argc, argv);

    

    /* initialize strings */
    true    = ajStrNew();
    cross   = ajStrNew();
    false   = ajStrNew(); 
    unknown = ajStrNew();   


    /* GET VALUES FROM ACD */
    hitsin       = ajAcdGetInfile("hitsin");
    datafile     = ajAcdGetOutfile("datafile");
    true    = ajAcdGetString("true");
    false   = ajAcdGetString("false");
    cross   = ajAcdGetString("cross");    
    unknown = ajAcdGetString("unknown");


    /* Check signature hits file */
    if(!hitsin)
        ajFatal("Could not open signature hits file\n");
    
    else
        ajFmtPrint("Signature hits file read ok\n");
    

    

    /* Call sigplot_CountHits */
    sigplot_CountHits(hitsin, &data);
    

    ajFmtPrint("Number of hits in hits file = %d\n", (data)->num_hits);


    /* Call sigplot_HitProportion */
    sigplot_HitProportion(hitsin, &data, &prob_array);
    

    /* Call sigplot_DataWrite */
    sigplot_DataWrite(datafile, data, prob_array, true, false, cross, unknown);
    
    

    /* Tidy up */
    sigplot_ScopdataDel(&data);
    ajFileClose(&hitsin);
    ajFileClose(&datafile);
    ajStrDel(&true);
    ajStrDel(&false);
    ajStrDel(&cross);    
    ajStrDel(&unknown);



    /* Return */
    ajExit();
    return 0;


}


/* @funcstatic sigplot_CountHits *****************************************
 **
 ** Read signature hits file and count the number of hits, by reading the 
 ** HI line and storing the value for the rank of the hit.
 **
 ** @param [r] hitsin   [AjPFile]      File pointer to signature hits file
 ** @param [w] data     [AjPScopdata*] Scopdata object pointer
 **
 ** @return [AjBool] True on succcess
 ** @@
 ******************************************************************************/
static AjBool   sigplot_CountHits(AjPFile hitsin, AjPScopdata *data)
{
    
    AjPStr   line           = NULL;     /* Line of text */    
    AjPStr   name           = NULL;     /* String for file name */
    ajint    num_hits       = 0;        /* Variable to hold no. of hits */


    /* initialize strings */
    name    = ajStrNew();
    line    = ajStrNew();    
    

    /* Check signature hits file */
    if(!hitsin)
        ajFatal("Could not open signature hits file\n");


 
    /* Read first line */
    while(ajFileReadLine(hitsin, &line) && !ajStrPrefixC(line,"//"))
    {
        /* Checking for scop family information */
        if(ajStrPrefixC(line,"XX"))
            continue;

        else if(ajStrPrefixC(line,"CL"))
            continue;

        else if(ajStrPrefixC(line,"FO"))
            continue;

        
        else if(ajStrPrefixC(line,"SF"))
            continue;

        else if(ajStrPrefixC(line,"FA"))
            continue;

        /* Start of loop to count number of hits */
        else if(ajStrPrefixC(line,"HI")) 
        {
            /* loop through HI lines */
            /* Read in and store the rank (no. of hits) */
            /* from the HI line */
            while(ajFileReadLine(hitsin, &line) && !ajStrPrefixC(line,"//"))        
                ajFmtScanS(line, "%*s %d", &num_hits);
        }
            
    }

    /* Allocate memory for Scopdata structure */
    (*data) = sigplot_ScopdataNew();


    /* Code to copy and modify signature hits file name */
    /* for use in naming output files                   */
    /* Copy name of signature hits file to string name  */
    ajStrAss(&name, ajFileGetName(hitsin));

    /* remove trailing '.hits' */
    ajStrSubstituteCC(&name, ".hits", "");



    /* Assign number of hits and file name into structure */
    (*data)->num_hits = num_hits;
    ajStrAss(&(*data)->file_name, name);
    


    /* tidy up */
    ajStrDel(&line);
    ajStrDel(&name);


    /* return */
    return ajTrue;
}






/* @funcstatic sigplot_HitProportion *****************************************
 **
 ** Read signature hits file and count the number of hits, by reading the 
 ** HI line and storing the value for the rank of the hit.
 **
 ** @param [r] hitsin     [AjPFile]      File pointer to signature hits file
 ** @param [r] data       [AjPScopdata*] Scopdata object pointer
 ** @param [w] prob_array [AjPFloat2d*]  Probability of four classifications
 ** 
 ** @return [AjBool] True on succcess
 ** @@
 ******************************************************************************/
static AjBool  sigplot_HitProportion(AjPFile hitsin, AjPScopdata *data,
				     AjPFloat2d *prob_array)
{


    AjPStr   line           = NULL;     /* Line of text */
    static   AjPStr class   = NULL;     /* string to hold class */
    static   AjPStr fold    = NULL;     /* string to hold fold */
    static   AjPStr super   = NULL;     /* string to hold super */
    static   AjPStr family  = NULL;     /* string to hold family */
    AjPStr   temp           = NULL;     /* temp variable for classification */
    ajint    x              = 0;
    ajint    y              = 0;
    ajint    true           = 0;        /* Counter for no. of true classifications */
    ajint    false          = 0;        /* Counter for no. of false classifications */
    ajint    cross          = 0;        /* Counter for no. of cross classifications */
    ajint    unknown        = 0;        /* Counter for no. of unknown classifications */
    ajint    rank           = 0;

    /* Check args */    
    if(!hitsin)
        return ajFalse;
    

    /* reset posinter to start of file */
    ajFileSeek(hitsin, 0, 0);



    /* Allocate memory for the seq_pos array */
    *prob_array = ajFloat2dNewL((*data)->num_hits);
    

    /* Set reserved size */
    for(x=0; x<4; x++)
        for(y=0;y<(*data)->num_hits;y++)
            ajFloat2dPut(prob_array, x, y, (ajint)0);
    
    
    /* assign strings */
    class   = ajStrNew();

    fold    = ajStrNew();
    super   = ajStrNew();
    family  = ajStrNew();
    line    = ajStrNew();
    temp    = ajStrNew();



    


    /* Read through the file */
    while(ajFileReadLine(hitsin, &line) && !ajStrPrefixC(line,"//"))
    {
        if(ajStrPrefixC(line,"DE"))
            continue;
        
        else if(ajStrPrefixC(line,"CL"))
            {
                ajStrAssC(&class,ajStrStr(line)+3);
                ajStrClean(&class);
            }
        else if(ajStrPrefixC(line,"FO"))
        {
            ajStrAssC(&fold,ajStrStr(line)+3);
            while((ajFileReadLine(hitsin,&line)))
            {
                if(ajStrPrefixC(line,"XX"))
                    break;
                ajStrAppC(&fold,ajStrStr(line)+3);
            }
            ajStrClean(&fold);
        }
        else if(ajStrPrefixC(line,"SF"))
        {
            ajStrAssC(&super,ajStrStr(line)+3);
            while((ajFileReadLine(hitsin,&line)))
            {
                if(ajStrPrefixC(line,"XX"))
                    break;
                ajStrAppC(&super,ajStrStr(line)+3);
            }
            ajStrClean(&super);
        }
        else if(ajStrPrefixC(line,"FA"))
        {
            ajStrAssC(&family,ajStrStr(line)+3);
            while((ajFileReadLine(hitsin,&line)))
            {
                if(ajStrPrefixC(line,"XX"))
                    break;
                ajStrAppC(&family,ajStrStr(line)+3);
            }
            ajStrClean(&family);
        }
        else if(ajStrPrefixC(line,"XX"))
            continue;

        else if(ajStrPrefixC(line,"HI"))
        {
            /* perform calculation on first line of HI */
            ajFmtScanS(line, "%*s %d %*s %*s %*s %S", &rank, &temp);          
            
            /* If classification = true, determine the                */
            /* proportion of each classification and write into array */
            if(ajStrPrefixC(temp, "TRUE"))
            {
                true++;
                ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));

            }
            
            /* If classification = false, determine the */
            /* proportion of each classification and write value into array */
            else if(ajStrPrefixC(temp, "FALSE"))
            {
                false++;
                ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));
            }
            
            /* If classification = cross, determine the */
            /* proportion of each classification and write value into array */
            else if(ajStrPrefixC(temp, "CROSS"))
            {
                cross++;
                ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));
            }
            
            /* If classification = unknown, determine the */
            /* proportion of each classification and write value into array */
            else if(ajStrPrefixC(temp, "UNKNOWN"))
            {
                unknown++;
                ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));
            }
            
            /* Start of loop for calculating */
            /* proportion of each classification */
            while(ajFileReadLine(hitsin, &line) && !ajStrPrefixC(line,"XX"))        
            {
                /* Scan rank and secondary classification into variables */
                ajFmtScanS(line, "%*s %d %*s %*s %*s %S", &rank, &temp);                

                /* If classification = true, determine the */
                /* proportion of each classification and write value into array */
                if(ajStrPrefixC(temp, "TRUE"))
                {
                    true++;
                    ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                    ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                    ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                    ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));
                }

                /* If classification = false, determine the */
                /* proportion of each classification and write value into array */
                else if(ajStrPrefixC(temp, "FALSE"))
                {
                    false++;
                    ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                    ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                    ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                    ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));

                }

                /* If classification = cross, determine the */
                /* proportion of each classification and write value into array */
                else if(ajStrPrefixC(temp, "CROSS"))
                {
                    cross++;
                    ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                    ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                    ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                    ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));
                }       

                /* If classification = unknown, determine the */
                /* proportion of each classification and write value into array */
                else if(ajStrPrefixC(temp, "UNKNOWN"))
                {
                    unknown++;  
                    ajFloat2dPut(prob_array, 0, rank, (float)((float)true/(float)rank));
                    ajFloat2dPut(prob_array, 1, rank, (float)((float)false/(float)rank));
                    ajFloat2dPut(prob_array, 2, rank, (float)((float)cross/(float)rank));
                    ajFloat2dPut(prob_array, 3, rank, (float)((float)unknown/(float)rank));
                }
            }
        }
    }


    /* Assign SCOP records */
    ajStrAss(&(*data)->Class,class);
    ajStrAss(&(*data)->Fold,fold);
    ajStrAss(&(*data)->Superfamily,super);
    ajStrAss(&(*data)->Family,family); 
       
    

    /* Tidy up */
    ajStrDel(&class);
    ajStrDel(&fold);
    ajStrDel(&class);
    ajStrDel(&family);
    ajStrDel(&line);
    ajStrDel(&temp);



    /* return */
    return ajTrue;
    
}






/* @funcstatic sigplot_DataWrite *****************************************
 **
 ** Read prob_array and Scopdata structure, and write a data file suitable
 ** for plotting using the application GNUPLOT
 **
 ** @param [w] datafile   [AjPFile]     File pointer to output datafile
 ** @param [r] data       [AjPScopdata] Scopdata object pointer
 ** @param [r] prob_array [AjPFloat2d]  Array of probability of four
 **                                     classifications
 ** @param [r] true       [AjPStr]    Name of sigplot data file for true hits
 ** @param [r] false      [AjPStr]    Name of sigplot data file for true hits
 ** @param [r] cross      [AjPStr]    Name of sigplot data file for true hits
 ** @param [r] unknown    [AjPStr]    Name of sigplot data file for true hits
 **
 ** @return [AjBool] True on succcess
 ** @@
 ******************************************************************************/
static AjBool  sigplot_DataWrite(AjPFile datafile, AjPScopdata data,
				 AjPFloat2d prob_array, 
				 AjPStr true, AjPStr false, AjPStr cross,
				 AjPStr unknown)
{
    ajint       x          = 0;     /* Loop counter */
    AjPFile     truePtr    = NULL;  /* Pointer to true data file */
    AjPFile     crossPtr   = NULL;  /* Pointer to cross data file */
    AjPFile     falsePtr   = NULL;  /* Pointer to false data file */
    AjPFile     unknownPtr = NULL;  /* Pointer to unknown data file */
        

    
    /* Assign file pointers */
    truePtr    = ajFileNewOut(true);
    crossPtr   = ajFileNewOut(cross);
    falsePtr   = ajFileNewOut(false);
    unknownPtr = ajFileNewOut(unknown);


    /* Print probabilities from array to data files */
    for(x = 0; x <(data)->num_hits; x++)
    {
        ajFmtPrintF(truePtr,    "%d    %4f\n", x, ajFloat2dGet(prob_array, 0, x));
        ajFmtPrintF(crossPtr,   "%d    %4f\n", x, ajFloat2dGet(prob_array, 1, x));
        ajFmtPrintF(falsePtr,   "%d    %4f\n", x, ajFloat2dGet(prob_array, 2, x));
        ajFmtPrintF(unknownPtr, "%d    %4f\n", x, ajFloat2dGet(prob_array, 3, x));
    }


    /* Write main .dat file */
    ajFmtPrintF(datafile,    "#  GNUPLOT data file suitable for plotting signature performance\n");
    ajFmtPrintF(datafile,    "#\n");
    ajFmtPrintF(datafile,    "#\n");
    ajFmtPrintF(datafile,    "#  %S\n", data->Class);
    ajFmtPrintF(datafile,    "#  XX\n");
    ajFmtPrintF(datafile,    "#  %S\n", data->Fold);
    ajFmtPrintF(datafile,    "#  XX\n");
    ajFmtPrintF(datafile,    "#  %S\n", data->Superfamily);
    ajFmtPrintF(datafile,    "#  XX\n");
    ajFmtPrintF(datafile,    "#  %S\n", data->Family);
    ajFmtPrintF(datafile,    "#  XX\n");
    ajFmtPrintF(datafile,    "set title \"Graph of Signature Performance\"\n");
    ajFmtPrintF(datafile,    "set xlabel \"Number of Hits\"\n");
    ajFmtPrintF(datafile,    "set ylabel \"Proportion of total hits\"\n");

    ajFmtPrintF(datafile,    "set nokey\n");
    ajFmtPrintF(datafile,    "set key top outside title \"Legend\" box 3 \n");
    ajFmtPrintF(datafile,    "set data style points\n");
    ajFmtPrintF(datafile,    "set pointsize 0.45\n");
    /* Execute plot command to plot data files */
    ajFmtPrintF(datafile,    "plot \"%S\" smooth bezier title \"True Hits\", \"%S\" smooth bezier title \"Cross Hits\", \"%S\" smooth bezier title \"False Hits\", \"%S\" smooth bezier title \"Unknown Hits\"\n", true, cross, false, unknown);
 
    
    /* Close files */
    ajFileClose(&truePtr);
    ajFileClose(&crossPtr);
    ajFileClose(&falsePtr);
    ajFileClose(&unknownPtr);


    
    /* Return */
    return ajTrue;
    

}




/* @funcstatic sigplot_ScopdataNew *******************************************
**
** Scopdata object constructor.
**
** 
** @return [AjPScopdata] Pointer to a Scopdata object
** @@
******************************************************************************/
static AjPScopdata  sigplot_ScopdataNew()
{
    AjPScopdata ret = NULL;
    


    AJNEW0(ret);
    ret->Class=ajStrNew();
    ret->Fold=ajStrNew();
    ret->Superfamily=ajStrNew();
    ret->Family=ajStrNew();
    ret->file_name=ajStrNew();

    return ret;
}


/* @funcstatic sigplot_ScopdataDel ********************************************
**
** Destructor for Scopdata object.
**
** @param [w] pthis [AjPScopdata*] Scopdata object pointer
**
** @return [void]
** @@
******************************************************************************/

static void sigplot_ScopdataDel(AjPScopdata *pthis)
{

    ajStrDel(&(*pthis)->Class);
    ajStrDel(&(*pthis)->Fold);
    ajStrDel(&(*pthis)->Superfamily);
    ajStrDel(&(*pthis)->Family);
    ajStrDel(&(*pthis)->file_name);

    
    AJFREE(*pthis);
    *pthis=NULL;
    
    /* return */
    return;
}




 
















